library ngpack;

import 'dart:convert';

import 'package:args/command_runner.dart';
import 'package:ngpack/src/text_decoder.dart';
import 'package:ngpack/src/tsv_reader.dart';
import 'package:universal_io/io.dart';
import 'package:glob/glob.dart';
import 'package:path/path.dart' as path;
import 'package:glob/list_local_fs.dart';
import 'package:ngpack/ngpack.dart';

abstract class KeyCommand extends Command {
  XorKey? key;
  String? pattern;
  String? filename;

  bool get checkInputExists => true;

  KeyCommand({String keyDefaultTo = 'auto'}) {
    argParser.addOption('key',
        help: 'Name of the key to decrypt/encrypt the data.\n'
            'Possible names: auto, 56ad, 5bad, 566d, 5b6d, delores\n',
        defaultsTo: keyDefaultTo,
        abbr: 'k');

    argParser.addOption('pattern',
        help: 'Pattern to use to match entries.', defaultsTo: '*', abbr: 'p');
  }

  @override
  void run() {
    key =
        argResults!['key'] == 'auto' ? null : knownXorKeys[argResults!['key']];
    pattern = argResults!['pattern'];

    if (argResults!.rest.isEmpty ||
        (checkInputExists && !File(argResults!.rest[0]).existsSync())) {
      throw UsageException('Please specify a ggpack file', usage);
    }

    filename = argResults!.rest[0];
  }
}

class ListCommand extends KeyCommand {
  @override
  final name = 'list';
  @override
  final description = 'List files in the ggpack matching the pattern.';

  ListCommand();

  @override
  void run() {
    super.run();
    final file = GGPackDecoder.fromFile(filename!, xorKey: key);
    file.where((e) => Glob(pattern!).matches(e.filename)).forEach(print);
  }
}

class ExtractCommand extends KeyCommand {
  @override
  final name = 'extract';
  @override
  final description =
      'Extract the files from the ggpack matching the pattern to the current working directory.';

  ExtractCommand() {
    argParser.addFlag('macro');
  }

  @override
  void run() async {
    super.run();

    final macroByTextId =
        argResults!['macro'] ? await _getMacroByTextId() : null;

    final file = GGPackDecoder.fromFile(filename!, xorKey: key);
    final entries =
        file.where((e) => Glob(pattern!).matches(e.filename)).toList();
    var i = 0;
    entries.forEach((e) {
      var progress = 100.0 * (++i) / entries.length;
      extractEntry(macroByTextId, file, e.filename, progress);
    });
    print('\r${entries.length} file(s) extracted'.padRight(72));
  }

  static void extractEntry(Map<int, String>? macroByTextId,
      GGPackDecoder decoder, String entry, double progress) {
    var outFile = File(entry);
    final data = decoder.extract(entry);
    if (macroByTextId != null && _useMacro(entry)) {
      final content = utf8.decode(data);
      outFile.writeAsStringSync(
          TextDecoder((id) => macroByTextId[id] ?? '"@$id"').convert(content));
    } else {
      outFile.writeAsBytesSync(data);
    }
    stdout.write(
        '\rExtracting ${entry.padRight(48)} ${progress.toStringAsFixed(1).padLeft(5)}%');
  }

  static bool _useMacro(String entry) {
    final ext = path.extension(entry);
    return ext == '.bnut' || ext == '.nut' || ext == '.wimpy';
  }

  static GGPackDecoder _getDecoderWithTsv(List<GGPackDecoder> decoders) {
    return decoders.firstWhere((decoder) =>
        decoder.any((element) => element.filename == 'ThimbleweedText_en.tsv'));
  }

  static GGPackDecoder _getDecoderWithLips(List<GGPackDecoder> decoders) {
    return decoders.firstWhere((decoder) =>
        decoder.any((element) => element.filename.endsWith('.lip')));
  }

  Future<Map<int, String>> _getMacroByTextId() async {
    var decoders = await _getDecoders();
    final decoder = _getDecoderWithTsv(decoders);
    final tsvContent = decoder.extractContent('ThimbleweedText_en.tsv');
    final textById = await TsvStreamReader().read(Stream.value(tsvContent));

    final entries = _getTextsByTextId(decoders);
    final result = textById.map((key, value) {
      stdout.write('\rProcessing $key...');
      final name = _getMacroName(entries[key]?.toSet() ?? <String>{});
      return MapEntry(key, '$name($key, "$value")');
    });
    print('\r'.padRight(64));
    return result;
  }

  static String _getMacroName(Set<String> names) {
    if (names.isEmpty) return 'NOTALK';
    if (names.length == 1) return names.first;
    if (names.length == 2 && names.contains('RAY') && names.contains('REYES')) {
      return 'AGENT';
    }
    var result = StringBuffer();
    if (names.contains('RAY') && names.contains('REYES')) {
      names..remove('RAY')..remove('REYES');
      result.write('A');
    }
    names.forEach((name) {
      result.write(name[0]);
    });
    return 'PLAYER_$result';
  }

  static Map<int, List<String>> _getTextsByTextId(
      List<GGPackDecoder> decoders) {
    final decoder = _getDecoderWithLips(decoders);
    // list all .lip files and get a map giving a List of strings for a given text id
    final regex = RegExp(r'(\w+)_(\d+)\.lip');
    return decoder
        .map((e) => e.filename)
        .where((element) => element.endsWith('.lip'))
        .fold(<int, List<String>>{}, (Map<int, List<String>> p, String a) {
      final m = regex.firstMatch(a);
      final name = m!.group(1);
      final id = int.parse(m.group(2)!);
      p.putIfAbsent(id, () => <String>[]);
      p[id]!.add(name!);
      return p;
    });
  }

  Future<List<GGPackDecoder>> _getDecoders() async {
    var decoders = <GGPackDecoder>[];
    for (var filename in argResults!.rest) {
      final bytes = await File(filename).readAsBytes();
      final decoder = GGPackDecoder(bytes, xorKey: key);
      decoders.add(decoder);
    }
    return decoders;
  }
}

class CreateCommand extends KeyCommand {
  @override
  final name = 'create';
  @override
  final description = 'Create a ggpack from the files matching the pattern.';
  @override
  final bool checkInputExists = false;

  CreateCommand() : super(keyDefaultTo: '56ad');

  @override
  Future<void> run() async {
    super.run();
    var builder =
        GGPackBuilder(key ?? knownXorKeys.fromId(KnownXorKeyId.Key56ad));
    var count = 0;
    await Glob(pattern!).list().where((e) => e is File).forEach((e) {
      count++;
      final entry = path.basename(e.path);
      stdout.write('\rAdding entry ${e.path.padRight(48)}');
      builder.addFile(entry, e.path);
    });
    File(filename!).writeAsBytesSync(builder.build(), flush: true);
    print('\r${filename!} created with $count entries.'.padRight(72));
  }
}

void main(List<String> arguments) {
  CommandRunner(
      'ngpack', 'A tool to list/extract/create Thimbleweed Park ggpack file.')
    ..addCommand(ListCommand())
    ..addCommand(ExtractCommand())
    ..addCommand(CreateCommand())
    ..run(arguments).catchError((error) {
      if (error is! UsageException) throw error;
      print(error);
      exit(64); // Exit code 64 indicates a usage error.
    });
}
